"Slots-related checks and inspection tools"
import platform
import sys
from typing import Collection, Iterator, Optional

try:
    from typing import is_typeddict
except ImportError:  # pragma: no cover
    from typing_extensions import is_typeddict


def slots(c: type) -> Optional[Collection[str]]:
    try:
        slots_raw = c.__dict__["__slots__"]
    except KeyError:
        return None
    if isinstance(slots_raw, str):
        return (slots_raw,)
    elif isinstance(slots_raw, Iterator):
        raise NotImplementedError("Iterator __slots__ not supported. See #22")
    else:
        return slots_raw


def has_slots(c: type) -> bool:
    return (
        "__slots__" in c.__dict__
        or not (issubclass(c, BaseException) or is_pure_python(c))
        or is_typeddict(c)
    )


def has_slotless_base(c: type) -> bool:
    return not all(map(has_slots, c.__bases__))


def slots_overlap(c: type) -> bool:
    maybe_slots = slots(c)
    if maybe_slots is None:
        return False
    slots_ = set(maybe_slots)
    for ancestor in c.__mro__[1:]:
        if not slots_.isdisjoint(slots(ancestor) or ()):
            return True
    return False


def has_duplicate_slots(c: type) -> bool:
    slots_ = slots(c) or ()
    return len(set(slots_)) != len(list(slots_))


# The 'is a pure python class' logic below is adapted
# from https://stackoverflow.com/a/41012823/

# If the active Python interpreter is the official CPython interpreter,
# prefer a more reliable CPython-specific solution guaranteed to succeed.
if platform.python_implementation() == "CPython":
    # Magic numbers defined by the Python codebase at "Include/object.h".
    Py_TPFLAGS_IMMUTABLETYPE = 1 << 8
    Py_TPFLAGS_HEAPTYPE = 1 << 9

    # Starting with CPython 3.10, `Py_TPFLAGS_HEAPTYPE` should no longer
    # be relied on, and `!Py_TPFLAGS_IMMUTABLETYPE` should be used instead.
    if sys.version_info >= (3, 10):

        def is_pure_python(cls: type) -> bool:
            "Whether the class is pure-Python or C-based"
            return not (cls.__flags__ & Py_TPFLAGS_IMMUTABLETYPE)

    else:  # pragma: no cover

        def is_pure_python(cls: type) -> bool:
            "Whether the class is pure-Python or C-based"
            return bool(cls.__flags__ & Py_TPFLAGS_HEAPTYPE)


# Else, fallback to a CPython-agnostic solution typically but *NOT*
# necessarily succeeding. For all real-world objects of interest, this is
# effectively successful. Edge cases exist but are suitably rare.
else:  # pragma: no cover

    def is_pure_python(cls: type) -> bool:
        "Whether the class is pure-Python or C-based"
        return "__dict__" in dir(cls) or hasattr(cls, "__slots__")



"Logic for gathering and managing the configuration settings"

import configparser
from dataclasses import dataclass, fields
from itertools import chain
from pathlib import Path
from typing import ClassVar, Collection, Mapping, Optional, Type, TypeVar

import tomli

from .common import add_slots, both

RegexStr = str
"A regex string in Python's verbose syntax"
DEFAULT_MODULE_EXCLUDE_RE = r"(^|\.)__main__(\.|$)"
_T = TypeVar("_T")


@add_slots
@dataclass(frozen=True)
class PartialConfig:
    "Configuration options where not all options may be defined."
    strict_imports: Optional[bool]
    require_subclass: Optional[bool]
    require_superclass: Optional[bool]
    include_modules: Optional[RegexStr]
    exclude_modules: Optional[RegexStr]
    include_classes: Optional[RegexStr]
    exclude_classes: Optional[RegexStr]

    EMPTY: ClassVar["PartialConfig"]

    @classmethod
    def load(cls, p: Path) -> "PartialConfig":
        "May raise various exceptions on parse problems. File must exist."
        if p.suffix == ".toml":
            return cls._load_toml(p)
        elif p.suffix in (".ini", ".cfg"):
            return cls._load_ini(p)
        else:
            raise ValueError(
                "Settings file with invalid extension. "
                "Expected .toml, .ini, or .cfg"
            )

    @classmethod
    def _load_toml(cls, p: Path) -> "PartialConfig":
        with p.open("rb") as rfile:
            return cls._load_confmap(
                tomli.load(rfile).get("tool", {}).get("slotscheck", {})
            )

    @classmethod
    def _load_ini(cls, p: Path) -> "PartialConfig":
        cfg = configparser.ConfigParser()
        cfg.read(p, encoding="utf-8")
        return cls._load_confmap(
            {
                k: cfg.BOOLEAN_STATES.get(v, v)
                for k, v in cfg.items("slotscheck")
            }
            if cfg.has_section("slotscheck")
            else {}
        )

    @classmethod
    def _load_confmap(cls, m: Mapping[str, object]) -> "PartialConfig":
        if not m.keys() <= _ALLOWED_KEYS.keys():
            raise InvalidKeys(m.keys() - _ALLOWED_KEYS)

        return PartialConfig(
            **{
                key.replace("-", "_"): _extract_value(m, key, expect_type)
                for key, expect_type in _ALLOWED_KEYS.items()
            },
        )


PartialConfig.EMPTY = PartialConfig(None, None, None, None, None, None, None)


@dataclass(frozen=True)
class Config(PartialConfig):
    "A full set of options"
    __slots__ = ()
    strict_imports: bool
    require_subclass: bool
    require_superclass: bool
    exclude_modules: RegexStr

    DEFAULT: ClassVar["Config"]

    def apply(self, other: PartialConfig) -> "Config":
        return Config(
            **{
                f.name: _none_or(getattr(other, f.name), getattr(self, f.name))
                for f in fields(PartialConfig)
            }
        )


Config.DEFAULT = Config(
    strict_imports=True,
    require_subclass=False,
    require_superclass=True,
    include_modules=None,
    exclude_modules=DEFAULT_MODULE_EXCLUDE_RE,
    include_classes=None,
    exclude_classes=None,
)


def collect(
    from_cli: PartialConfig, cwd: Path, config: Optional[Path]
) -> Config:
    "Gather and combine configuration options from the available sources"
    confpath = config or find_config_file(cwd)
    conf = PartialConfig.load(confpath) if confpath else PartialConfig.EMPTY
    return Config.DEFAULT.apply(conf).apply(from_cli)


_CONFIG_FILENAMES = ("pyproject.toml", "setup.cfg")


def find_config_file(path: Path) -> Optional[Path]:
    "Find a configuration file in the given directory or its parents"
    candidates = (
        directory / name
        for directory in chain((path,), path.parents)
        for name in _CONFIG_FILENAMES
    )
    return next(
        filter(
            both(
                Path.is_file,  # type: ignore[arg-type]
                _has_slotscheck_section,
            ),
            candidates,
        ),
        None,
    )


def _has_slotscheck_section(p: Path) -> bool:
    with p.open("rb") as rfile:
        if p.suffix == ".toml":
            return "slotscheck" in tomli.load(rfile).get("tool", {})
        else:
            assert p.suffix == ".cfg"
            cfg = configparser.ConfigParser()
            cfg.read(p, encoding="utf-8")
            return cfg.has_section("slotscheck")


class InvalidKeys(Exception):
    def __init__(self, keys: Collection[str]) -> None:
        self.keys = keys

    def __str__(self) -> str:
        return (
            "Invalid configuration key(s): "
            + ", ".join(map("'{}'".format, sorted(self.keys)))
            + "."
        )


class InvalidValueType(Exception):
    def __init__(self, key: str) -> None:
        self.key = key

    def __str__(self) -> str:
        return f"Invalid value type for '{self.key}'."


def _none_or(a: Optional[_T], b: _T) -> _T:
    return b if a is None else a


def _extract_value(
    c: Mapping[str, object], key: str, expect_type: Type[_T]
) -> Optional[_T]:
    try:
        raw_value = c[key]
    except KeyError:
        return None
    else:
        if isinstance(raw_value, expect_type):
            return raw_value
        else:
            raise InvalidValueType(key)


_ALLOWED_KEYS = {
    "strict-imports": bool,
    "require-subclass": bool,
    "require-superclass": bool,
    "include-modules": str,
    "exclude-modules": str,
    "include-classes": str,
    "exclude-classes": str,
}



import re
import sys
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass
from functools import partial
from itertools import chain, filterfalse, starmap
from operator import attrgetter, itemgetter, methodcaller, not_
from pathlib import Path
from textwrap import indent
from typing import (
    Collection,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
)

import click

from . import config
from .checks import (
    has_duplicate_slots,
    has_slotless_base,
    has_slots,
    is_pure_python,
    slots,
    slots_overlap,
)
from .common import (
    Predicate,
    add_slots,
    both,
    compose,
    flatten,
    groupby,
    map_optional,
)
from .discovery import (
    AbsPath,
    FailedImport,
    ModuleLocated,
    ModuleName,
    ModuleTree,
    UnexpectedImportLocation,
    consolidate,
    find_modules,
    module_tree,
    walk_classes,
)


@click.command("slotscheck")
@click.argument(
    "FILES",
    type=click.Path(path_type=Path, exists=True, resolve_path=True),
    required=False,
    nargs=-1,
)
@click.option(
    "-m",
    "--module",
    help="Check this module. Cannot be combined with FILES argument. "
    "Can be repeated multiple times to scan several modules. ",
    multiple=True,
)
@click.option(
    "--require-superclass/--no-require-superclass",
    help="Report an error when a slots class inherits from "
    "a non-slotted class.",
    default=None,
    show_default="required",
)
@click.option(
    "--require-subclass/--no-require-subclass",
    help="Report an error when a non-slotted class inherits from "
    "a slotted class. In effect, this option enforces the use of slots "
    "wherever possible.",
    default=None,
    show_default="not required",
)
@click.option(
    "--include-modules",
    help="A regular expression that matches modules to include. "
    "Exclusions are determined first, then inclusions. "
    "Uses Python's verbose regex dialect, so whitespace is mostly ignored.",
)
@click.option(
    "--exclude-modules",
    help="A regular expression that matches modules to exclude. "
    "Excluded modules will not be imported. "
    "Uses Python's verbose regex dialect, so whitespace is mostly ignored.",
    show_default=f"``{config.DEFAULT_MODULE_EXCLUDE_RE}``",
)
@click.option(
    "--include-classes",
    help="A regular expression that matches classes to include. "
    "Use ``:`` to separate module and class paths. "
    "For example: ``app\\.config:.*Settings``, ``:(Foo|Bar)``. "
    "Exclusions are determined first, then inclusions. "
    "Uses Python's verbose regex dialect, so whitespace is mostly ignored.",
)
@click.option(
    "--exclude-classes",
    help="A regular expression that matches classes to exclude. "
    "Use ``:`` to separate module and class paths. "
    "For example: ``app\\.config:Settings``, ``:.*(Exception|Error)``. "
    "Uses Python's verbose regex dialect, so whitespace is mostly ignored.",
    show_default="``^$``",
)
@click.option(
    "--strict-imports/--no-strict-imports",
    help="Treat failed imports as errors.",
    default=None,
    show_default="strict",
)
@click.option(
    "-v", "--verbose", is_flag=True, help="Display extra descriptive output."
)
@click.option(
    "--settings",
    help="Path to the configuration file to use. "
    "Allowed extensions are toml, cfg, ini.",
    type=click.Path(
        path_type=Path, exists=True, resolve_path=True, dir_okay=False
    ),
)
@click.version_option()
def root(
    files: Sequence[AbsPath],
    module: Sequence[str],
    verbose: bool,
    settings: Optional[AbsPath],
    require_superclass: Optional[bool],
    require_subclass: Optional[bool],
    strict_imports: Optional[bool],
    exclude_classes: Optional[config.RegexStr],
    include_classes: Optional[config.RegexStr],
    exclude_modules: Optional[config.RegexStr],
    include_modules: Optional[config.RegexStr],
) -> None:
    "Check whether your __slots__ are working properly."
    conf = config.collect(
        config.PartialConfig(
            strict_imports=strict_imports,
            require_superclass=require_superclass,
            require_subclass=require_subclass,
            exclude_classes=exclude_classes,
            include_classes=include_classes,
            exclude_modules=exclude_modules,
            include_modules=include_modules,
        ),
        Path.cwd(),
        settings,
    )
    if not (files or module):
        print("No files or modules given. Nothing to do!")
        exit(0)

    try:
        classes, modules = _collect(files, module, conf)
    except ModuleNotFoundError as e:
        print(
            f"ERROR: Module '{e.name}' not found.\n\n"
            "See slotscheck.rtfd.io/en/latest/discovery.html\n"
            "for help resolving common import problems."
        )
        exit(1)
    except UnexpectedImportLocation as e:
        print(
            """\
Cannot check due to import ambiguity.
The given files do not correspond with what would be imported:

  'import {0.module}' would load from:
  {0.actual}
  instead of:
  {0.expected}

You may need to define $PYTHONPATH or run as 'python -m slotscheck'
to ensure the correct files can be imported.

See slotscheck.rtfd.io/en/latest/discovery.html
for more information on why this happens and how to resolve it.""".format(
                e
            )
        )
        exit(1)

    if not (modules.filtered or modules.skipped):
        print(
            "Files or modules given, but filtered out by exclude/include. "
            "Nothing to do!"
        )
        exit(0)
    messages = list(
        chain(
            map(
                partial(Message, error=conf.strict_imports),
                sorted(modules.skipped, key=attrgetter("failure.module")),
            ),
            _check_classes(
                classes,
                conf.require_superclass,
                conf.include_classes,
                conf.exclude_classes,
                conf.require_subclass,
            ),
        )
    )
    for msg in messages:
        print(msg.for_display(verbose))

    any_errors = any(m.error for m in messages)

    if any_errors:
        print("Oh no, found some problems!")
    else:
        print("All OK!")

    if verbose:
        _print_report(
            modules,
            classes,
        )
    else:
        print(
            f"Scanned {sum(map(len, modules.filtered))} module(s), "
            f"{len(classes)} class(es)."
        )

    if any_errors:
        exit(1)


class Notice(ABC):
    "Base class for notices to be displayed"
    __slots__ = ()

    @abstractmethod
    def for_display(self, verbose: bool) -> str:
        raise NotImplementedError()


@add_slots
@dataclass(frozen=True)
class ModuleSkipped(Notice):
    "Notice that a module has been skipped due to an import failure."
    failure: FailedImport

    def for_display(self, verbose: bool) -> str:
        return (
            f"Failed to import '{self.failure.module}'."
            + verbose * f"\nDue to {self.failure.exc!r}"
        )


@add_slots
@dataclass(frozen=True)
class OverlappingSlots(Notice):
    "Notice that slots on a class overlap with its superclass(es)."
    cls: type

    def for_display(self, verbose: bool) -> str:
        return (
            f"'{_class_fullname(self.cls)}' defines overlapping slots."
            + verbose
            * (
                "\nSlots already defined in superclass:\n"
                + _bulletlist(
                    f"'{name}' ({_class_fullname(base)})"
                    for name, base in sorted(
                        _overlapping_slots(self.cls), key=itemgetter(0)
                    )
                )
            )
        )


def _overlapping_slots(c: type) -> Iterable[Tuple[str, type]]:
    maybe_slots = slots(c)
    assert maybe_slots is not None
    slots_ = set(maybe_slots)
    for base in c.mro()[1:]:
        for overlap in slots_.intersection(slots(base) or ()):
            yield (overlap, base)


@add_slots
@dataclass(frozen=True)
class DuplicateSlots(Notice):
    "Notice that a class defines duplicate slots."
    cls: type

    def for_display(self, verbose: bool) -> str:
        return f"'{_class_fullname(self.cls)}' has duplicate slots." + (
            verbose
            * (
                "\nDuplicate slot names:\n"
                + _bulletlist(map("'{}'".format, _duplicate_slots(self.cls)))
            )
        )


def _duplicate_slots(c: type) -> Iterable[str]:
    return (
        slot for slot, count in Counter(slots(c) or ()).items() if count > 1
    )


@add_slots
@dataclass(frozen=True)
class BadSlotInheritance(Notice):
    "Notice that a class has slots, but some of its superclasses do not."
    cls: type

    def for_display(self, verbose: bool) -> str:
        return (
            f"'{_class_fullname(self.cls)}' has slots "
            "but superclass does not."
            + verbose
            * (
                "\nSuperclasses without slots:\n"
                + _bulletlist(
                    map(
                        compose(
                            "'{}'".format,
                            _class_fullname,
                        ),
                        _slotless_superclasses(self.cls),
                    )
                )
            )
        )


@add_slots
@dataclass(frozen=True)
class ShouldHaveSlots(Notice):
    "Notice that a class should have slots, but doesn't."
    cls: type

    def for_display(self, verbose: bool) -> str:
        return (
            f"'{_class_fullname(self.cls)}' has no slots, but it could have."
            + verbose
            * "\nThis is flagged because the strict `require-subclass` "
            "setting is active."
        )


@add_slots
@dataclass(frozen=True)
class Message:
    "A notice with error level."
    notice: Notice
    error: bool

    def for_display(self, verbose: bool) -> str:
        return (_format_error if self.error else _format_note)(
            self.notice.for_display(verbose)
        )


@add_slots
@dataclass(frozen=True)
class ModulesReport:
    "Report with data on modules excluded or skipped."
    all: Collection[ModuleTree]
    filtered: Collection[ModuleTree]
    skipped: Collection[ModuleSkipped]


def _collect(
    files: Collection[AbsPath],
    modules: Collection[ModuleName],
    conf: config.Config,
) -> Tuple[Collection[type], ModulesReport]:
    modulefilter = _create_filter(conf.include_modules, conf.exclude_modules)
    modules_inspected = list(
        starmap(
            module_tree,
            filter(
                compose(modulefilter, attrgetter("name")),
                _as_modules(files, modules),
            ),
        )
    )
    modules_all = consolidate(
        m for m in modules_inspected if not isinstance(m, FailedImport)
    )
    modules_filtered: Collection[ModuleTree] = list(
        map_optional(methodcaller("filtername", modulefilter), modules_all)
    )
    classes, modules_skipped = _collect_classes(modules_filtered)
    return classes, ModulesReport(
        modules_all,
        modules_filtered,
        list(
            chain(
                (
                    ModuleSkipped(m)
                    for m in modules_inspected
                    if isinstance(m, FailedImport)
                ),
                modules_skipped,
            )
        ),
    )


def _create_filter(include: Optional[str], exclude: str) -> Predicate[str]:
    excluder: Predicate[str] = compose(
        not_, re.compile(exclude, flags=re.VERBOSE).search
    )
    return (
        excluder
        if include is None
        else both(excluder, re.compile(include, flags=re.VERBOSE).search)
    )


def _as_modules(
    files: Collection[AbsPath], names: Iterable[ModuleName]
) -> Iterable[ModuleLocated]:
    if files and names:
        print(
            _format_error(
                "Specify either FILES argument or `-m/--module` "
                "option, not both."
            ),
            file=sys.stderr,
        )
        exit(2)
    elif files:
        return flatten(map(find_modules, files))
    else:
        return map(partial(ModuleLocated, expected_location=None), names)


def _collect_classes(
    trees: Collection[ModuleTree],
) -> Tuple[Collection[type], Collection[ModuleSkipped]]:
    classes: List[type] = []
    skipped: List[ModuleSkipped] = []
    for result in flatten(map(walk_classes, trees)):
        if isinstance(result, FailedImport):
            skipped.append(ModuleSkipped(result))
        else:
            classes.extend(result)
    return classes, skipped


def _print_report(
    modules: ModulesReport,
    classes: Collection[type],
) -> None:
    classes_by_status = groupby(
        classes,
        key=lambda c: None
        if not is_pure_python(c)
        else True
        if has_slots(c)
        else False,
    )
    print(
        """\
stats:
  modules:     {}
    checked:   {}
    excluded:  {}
    skipped:   {}

  classes:     {}
    has slots: {}
    no slots:  {}
    n/a:       {}""".format(
            sum(map(len, modules.all)),
            sum(map(len, modules.filtered)),
            sum(map(len, modules.all)) - sum(map(len, modules.filtered)),
            len(modules.skipped),
            len(classes),
            len(classes_by_status[True]),
            len(classes_by_status[False]),
            len(classes_by_status[None]),
        ),
        file=sys.stderr,
    )


def _check_classes(
    classes: Iterable[type],
    require_superclass: bool,
    include: Optional[str],
    exclude: Optional[str],
    require_subclass: bool,
) -> Iterator[Message]:
    return map(
        partial(Message, error=True),
        flatten(
            map(
                partial(
                    slot_messages,
                    require_subclass=require_subclass,
                    require_superclass=require_superclass,
                ),
                sorted(
                    _class_includes(
                        _class_excludes(classes, exclude),
                        include,
                    ),
                    key=_class_fullname,
                ),
            )
        ),
    )


def _class_excludes(
    classes: Iterable[type], exclude: Optional[str]
) -> Iterable[type]:
    return (
        filter(
            compose(
                not_,
                re.compile(exclude, flags=re.VERBOSE).search,
                _class_fullname,
            ),
            classes,
        )
        if exclude
        else classes
    )


def _class_includes(
    classes: Iterable[type], include: Optional[str]
) -> Iterable[type]:
    return (
        filter(
            compose(
                re.compile(include, flags=re.VERBOSE).search,
                _class_fullname,
            ),
            classes,
        )
        if include
        else classes
    )


def slot_messages(
    c: type, require_superclass: bool, require_subclass: bool
) -> Iterable[Notice]:
    if slots_overlap(c):
        yield OverlappingSlots(c)
    if has_duplicate_slots(c):
        yield DuplicateSlots(c)
    if require_superclass and has_slots(c) and has_slotless_base(c):
        yield BadSlotInheritance(c)
    elif require_subclass and not has_slots(c) and not has_slotless_base(c):
        yield ShouldHaveSlots(c)


_ERROR_PREFIX = "ERROR: "
_NOTE_PREFIX = "NOTE:  "


def _format_error(msg: str) -> str:
    return (
        _ERROR_PREFIX
        + indent(msg, " " * len(_ERROR_PREFIX))[len(_ERROR_PREFIX) :]  # noqa
    )


def _format_note(msg: str) -> str:
    return (
        _NOTE_PREFIX
        + indent(msg, " " * len(_NOTE_PREFIX))[len(_NOTE_PREFIX) :]  # noqa
    )


def _class_fullname(k: type) -> str:
    return f"{k.__module__}:{k.__qualname__}"


def _bulletlist(s: Iterable[str]) -> str:
    return "\n".join(map("- {}".format, s))


def _slotless_superclasses(c: type) -> Iterable[type]:
    return filterfalse(has_slots, c.mro())



"Tools to discover and inspect modules, packages, and classes"

import importlib
import pkgutil
from dataclasses import dataclass, field, replace
from functools import partial, reduce
from importlib._bootstrap_external import (  # type: ignore[import]
    _NamespaceLoader,
)
from importlib.abc import FileLoader
from importlib.machinery import ExtensionFileLoader
from inspect import isclass
from itertools import chain, takewhile
from pathlib import Path
from textwrap import indent
from types import ModuleType
from typing import (
    Callable,
    Collection,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Union,
)

from .common import add_slots, flatten, unique

ModuleName = str
"The full, dotted name of a module"

ModuleNamePart = str
"Part of a module name -- no dots"

ModuleTree = Union["Module", "Package"]

AbsPath = Path
"A resolved filepath. Contains no '..'."


def consolidate(trees: Iterable[ModuleTree]) -> Collection[ModuleTree]:
    "Deduplicate and merge module trees"
    seen: Dict[ModuleNamePart, ModuleTree] = {}
    for t in trees:
        try:
            overlap = seen[t.name]
        except KeyError:
            seen[t.name] = t
        else:
            seen[t.name] = overlap.merge(t)

    return seen.values()


@add_slots
@dataclass(frozen=True)
class Module:
    name: ModuleNamePart

    def __post_init__(self) -> None:
        assert "." not in self.name

    def display(self) -> str:
        return self.name

    def __iter__(self) -> Iterator[ModuleTree]:
        yield self

    def __len__(self) -> int:
        return 1

    def filtername(
        self, __pred: Callable[[ModuleName], bool], *, prefix: str = ""
    ) -> Optional[ModuleTree]:
        return self if __pred(prefix + self.name) else None

    def merge(self, other: ModuleTree) -> ModuleTree:
        "Merge along shared components. Raises if no shared components."
        if other.name == self.name:
            return other
        raise ValueError("Cannot merge modules without shared components.")


@add_slots
@dataclass(frozen=True)
class Package:
    name: ModuleNamePart
    content: FrozenSet[ModuleTree]

    def __post_init__(self) -> None:
        assert "." not in self.name

    def display(self) -> str:
        return (
            f"{self.name}"
            + ("\n" * bool(self.content))
            + "\n".join(
                indent(node.display(), " ")
                for node in sorted(
                    self.content, key=lambda n: (type(n) is Package, n.name)
                )
            )
        )

    def __iter__(self) -> Iterator[ModuleTree]:
        yield self
        yield from flatten(self.content)

    def __len__(self) -> int:
        return 1 + sum(map(len, self.content))

    def filtername(
        self, __pred: Callable[[ModuleName], bool], *, prefix: str = ""
    ) -> Optional[ModuleTree]:
        if not __pred(prefix + self.name):
            return None

        new_prefix = f"{prefix}{self.name}."
        return replace(
            self,
            content=frozenset(
                filter(
                    None,
                    (
                        sub.filtername(__pred, prefix=new_prefix)
                        for sub in self.content
                    ),
                )
            ),
        )

    def merge(self, other: ModuleTree) -> ModuleTree:
        "Merge along shared components. Raises if no shared components."
        if self.name != other.name:
            raise ValueError("Cannot merge modules without shared components.")
        if isinstance(other, Module):
            return self
        else:
            return Package(
                self.name,
                frozenset(consolidate(chain(self.content, other.content))),
            )


@dataclass(frozen=True)
class UnexpectedImportLocation(Exception):
    module: ModuleName
    expected: AbsPath
    actual: AbsPath


@add_slots
@dataclass(frozen=True)
class FailedImport:
    module: ModuleName
    exc: BaseException = field(compare=False, hash=False)


def module_tree(
    module: ModuleName,
    expected_location: Optional[AbsPath],
) -> Union[ModuleTree, FailedImport]:
    """May raise ModuleNotFoundError or UnexpectedImportLocation"""
    try:
        loader = pkgutil.get_loader(module)
    except BaseException as e:
        return FailedImport(module, e)
    *namespaces, name = module.split(".")
    location: AbsPath
    tree: ModuleTree
    if isinstance(loader, (FileLoader, ExtensionFileLoader)):
        assert isinstance(loader.path, str)  # type: ignore[union-attr]
        location = Path(loader.path)  # type: ignore[union-attr]
        tree = (
            _package(name, location.parent)
            if loader.is_package(module)
            else Module(name)
        )
    elif isinstance(loader, _NamespaceLoader):
        assert len(loader._path._path) == 1
        location = Path(loader._path._path[0])
        tree = _package(name, location)
    elif loader is None:
        raise ModuleNotFoundError(f"No module named '{module}'", name=module)
    elif module == "builtins":
        return Module(module)
    else:
        raise NotImplementedError(f"Unsupported module loader type: {loader}")

    if expected_location and location != expected_location:
        raise UnexpectedImportLocation(module, expected_location, location)

    return reduce(_add_namespace, reversed(namespaces), tree)


def _add_namespace(tree: ModuleTree, name: ModuleNamePart) -> ModuleTree:
    return Package(name, frozenset([tree]))


def _submodule(m: pkgutil.ModuleInfo) -> ModuleTree:
    if m.ispkg:
        [subdir] = m.module_finder.find_spec(
            m.name  # type: ignore
        ).submodule_search_locations
        return _package(m.name, Path(subdir))
    else:
        return Module(m.name)


def _is_submodule(m: pkgutil.ModuleInfo, path: AbsPath) -> bool:
    return getattr(m.module_finder, "path", "").startswith(str(path))


def _package(name: ModuleNamePart, path: AbsPath) -> Package:
    return Package(
        name,
        frozenset(
            map(
                _submodule,
                filter(
                    partial(_is_submodule, path=path),
                    pkgutil.walk_packages([str(path)]),
                ),
            )
        ),
    )


def walk_classes(
    n: ModuleTree, prefix: str = ""
) -> Iterator[Union[FailedImport, FrozenSet[type]]]:
    fullname = prefix + n.name
    try:
        module = importlib.import_module(fullname)
    except BaseException as e:
        # Sometimes packages make it impossible to import
        # certain modules directly or out of order, or without
        # some system or dependency requirement.
        # The exceptions can be quite exotic,
        # inheriting from BaseException in some cases!
        if isinstance(e, KeyboardInterrupt):
            raise
        yield FailedImport(fullname, e)
    else:
        yield frozenset(_classes_in_module(module))
        if isinstance(n, Package):
            yield from flatten(
                map(partial(walk_classes, prefix=fullname + "."), n.content)
            )


def _classes_in_module(module: ModuleType) -> Iterable[type]:
    return flatten(
        map(
            _walk_nested_classes,
            unique(
                filter(
                    partial(_is_module_class, module=module),
                    module.__dict__.values(),
                )
            ),
        )
    )


def _is_module_class(obj: object, module: ModuleType) -> bool:
    try:
        return isclass(obj) and obj.__module__ == module.__name__
    except Exception:
        # Some rare objects crash on introspection. It's best to exclude them.
        return False


def _walk_nested_classes(c: type) -> Iterable[type]:
    yield c
    yield from flatten(map(_walk_nested_classes, _nested_classes(c)))


def _nested_classes(c: type) -> Iterable[type]:
    return unique(
        filter(partial(_is_nested_class, parent=c), c.__dict__.values())
    )


def _is_nested_class(obj: object, parent: type) -> bool:
    try:
        return (
            isclass(obj)
            and obj.__module__ == parent.__module__
            and obj.__qualname__.startswith(parent.__name__ + ".")
        )
    except Exception:
        # Some rare objects crash on introspection. It's best to exclude them.
        return False


_INIT_PY = "__init__.py"


class ModuleLocated(NamedTuple):
    name: ModuleName
    expected_location: Optional[AbsPath]


def _is_module(p: AbsPath) -> bool:
    return (p.is_file() and p.suffixes == [".py"]) or _is_package(p)


def _is_package(p: AbsPath) -> bool:
    return p.is_dir() and (p / _INIT_PY).is_file()


def find_modules(p: AbsPath) -> Iterable[ModuleLocated]:
    "Recursively find modules at given path. Nonexistent Path is ignored"
    if p.name == _INIT_PY:
        yield from find_modules(p.parent)
    elif _is_module(p):
        parents = [p] + list(takewhile(_is_package, p.parents))
        yield ModuleLocated(
            ".".join(p.stem for p in reversed(parents)),
            (p / "__init__.py" if _is_package(p) else p),
        )
    elif p.is_dir():
        yield from flatten(map(find_modules, p.iterdir()))
